#ifndef _Fl_Sharable_Image_H_
#define _Fl_Sharable_Image_H_
#include <FL/Fl_Image.H>
#include <FL/Fl_Shared_Image.H> // This is necessary only to use typedef Fl_Shared_Handler



// Note: This class could be made aven ABI compatible with Fl_Shared_Image
// However because it will not be a part befire fltk-1.3, the api is rathed

class FL_EXPORT  Fl_Shareable_Image: public Fl_Image {

  //friend class Fl_JPEG_Image;
  //friend class Fl_PNG_Image;
  friend class Fl_Registered_Image;
  void update_from_registered(Fl_Image * r);
  friend int fl_image_is_shared(const Fl_Image *im);
  int base_count() const {return Fl_Image::count();}

protected:
  // Following should be rather part of Registered_Image class but are here only for backward compatibility
  // because these are protected so user have theoretical acces to them if he subclasses Fl_Shared_Image (which he should not)
  // If breaking this theoretical compatibility is acceptable, move those to Registered_Image class within .cxx
  // static Fl_Shareable_Image **images_;	// Shared (actualu registered) images
  // static int	num_images_;		// Number of registered images
  // static int	alloc_images_;		// Allocated registered images
  static Fl_Shared_Handler *handlers_;	// Additional format handlers
  static int	num_handlers_;		// Number of format handlers
  static int	alloc_handlers_;	// Allocated format handlers

  // Also most of the following members are here for compatibility only,
  // they could be removed to decrease size of the shell as they are duplicates of the internal image
  //int		original_;		// Original image? - slightly different meaning now...
  //int		refcount_;		// Number of times this image has been used, should be removed (and added only to Registered_Image) if binary compatibility is not necessary
  Fl_Image	*image_;		// Pointer to registered image
  //int		alloc_image_;		// Was the image allocated?  - should be removed (and added only to Registered_Image) if binary compatibility is not necessary

  //static int	compare(Fl_Shareable_Image **i0, Fl_Shareable_Image **i1);

  // Use get() and release() to load/delete images in memory...
  Fl_Shareable_Image();
  Fl_Shareable_Image(const char *n, Fl_Image *img = 0);
  void init();
  void assure_data() const;
public:
  //Destructor now is puplic.
  virtual ~Fl_Shareable_Image();
protected:
  // Deprecated, for mackward compatibility if you use in conslunction eith protected constructor
  // Fl_Shareable_Image(const char * name, int W = 0, int H = 0)
  void add();
  // For backward compatibility only.
  void update();

public:
  // This is redefined because the data might not yet exist at a time you want to access them.
  // Note that this function only "shadows" previous definition. For full compatibility
  // this function should be declared as virtual in the base class Fl_Image.
  const char * const *data() const {assure_data(); return image_? image_->data() : 0;}
  int count() const {assure_data();  return image_? image_->count() : 0;}

  /** Returns the filename of the shared image */
  // RK: This is  const-corrected from original Fl_Shared_Image
  const char  *name() const;

  /** Returns the number of references of this shared image. When reference is below 1, the image is deleted. */

  int refcount();

  // Deprecated but functional. Use operator delete instead.
  void release();


  void reload();
  // Cheap
  virtual Fl_Image *copy(int W, int H);

  // Copy - always cheap
  Fl_Image *copy() { return copy(w(), h()); }

  // Cheap
  virtual void color_average(Fl_Color c, float i);

  // Cheap
  virtual void desaturate();

  virtual void draw(int X, int Y, int W, int H, int cx, int cy);

  // This is not required because it shadows the same in Fl_Image and has the same implementation.
  // However removal would break ABI if somebody would dig a pointer to it (linker error undefined reference to...)
  void draw(int X, int Y) { draw(X, Y, w(), h(), 0, 0); }

  virtual void uncache();

  static Fl_Shareable_Image *find(const char *n, int W = 0, int H = 0);

  // Deprecated but fully functional, use operator new with constructor instead.
  // Obtained image can be both release()d or deleted.
  static Fl_Shareable_Image *get(const char *n, int W = 0, int H = 0);

  static Fl_Shareable_Image **images();
  static int		num_images();
  static void		add_handler(Fl_Shared_Handler f);
  static void		remove_handler(Fl_Shared_Handler f);



   ///////////////////  New functions  /////////////////////

  // Basic constructor: In conjunction with operator new it is equivalent to get() like:
  //   Fl_Shareable_Image::get(name, W, H);
  //   new Fl_Shareable_Image(name, W, H);
  Fl_Shareable_Image(const char * name, int W = 0, int H = 0);

  // A constructor which makes existing image sharable under particular name.
  // If the name is NULL, it assigns unique internall name to it which can be inspected by name().
  // Copies of such image will still share the same data as long as they have the same
  // W, H and desaturate() and collor_average() combination results to the same effect.
  // Parameter \a take_over tells Fl_Shareable_Image engine if it should destroy image when not
  // needed. In such a case \a image should be created on the heap (using operator new) eg
  //   Fl_Shareable_Image(new Fl_RGB_Image(...), 0, 1).
  // It does not say if it will destruct internal data of the \a image, this depends how
  // particular \a image was constructed itself. It also deprecates
  //   Fl_JPEG_Image(const char *name, const unsigned char *data)
  //   Fl_PNG_Image(const char * name, const unsigned char * buffer, int maxsize)
  // although those could be made fully functional for backward compatibility.
  // Note that it might uncache() the \a image (for now it uncaches if it is not
  // Fl_Shareable_Image but in the future it might not).
  // If image is shared image, and the name is not null it registers the image
  // under second name. In such a case if take_over is 1, it will release the original
  // when not needed anymore.
  // If the image is Fl_Shareable_Image and the name is null, it is equivalent to copy constructor
  Fl_Shareable_Image(Fl_Image * image, const char * name = 0, int take_over = 0);


  // Copy constructor. Cheap.
  Fl_Shareable_Image(const Fl_Shareable_Image &);


  // Assignment operator. Cheap.
  Fl_Shareable_Image & operator = (const Fl_Shareable_Image &);

  //  Resizes image. Sequential copies or resizes do not cause quality loss because images are always
  //  resized from original. If W is 0 it resizes to full-scale original. Cheap with assure_data_mode(0).
  void resize(int W = 0, int H = 0);


  // This assures data for particular image  - that is until it is resized (and If default policy OFF)
  // You can call assure_data() before accessing data(), data_w(), data_h().
  //   image->assure_data();
  //   const char * const * data = image->data();
  // However the data is valid only as long as there was not call to resize() or there was no call to new
  // function fl_draw(Fl_Shareable_Image *) as such a call might re-build internal image depending on transformation
  // matrix and graphics state. See also data(), data_w() and data_h().


  // w() and h() might not be synchronised with "current pixel data" sizes.  This might happen ONLY if you draw
  // an image using  fl_draw(Fl_Shareable_Image *, ...) function which obeys transdormation matrix: internal image might
  // have different size than encapsulating shell.  For instance if you have
  //
  //   Fl_Shared_Image * im1 = new Sl_Shared_Image("/some/image.jpg");
  //   Fl_Shared_Image * im2 = (Fl_Shared_Image *) im1->copy();
  //
  // and then in your drawing code you have:
  //
  //   fl_draw(im1, 0,0);
  //   fl_push_matrix();
  //   fl_scale(2,2);
  //   fl_draw(im2, 0, 0);
  //   fl_pop_matrix();
  //
  // both images will still have the same w() and h() but data_w(), data_h() for im2 will be twice than those of im1
  // because the internal image of im2 was resampled due to scalling factor.
  // Note that we made copy of the image for efficiency: if you draw an image at two different places with two
  // different scalling factors, the image does not need to be resampled each time the zooming changes.
  // If you wanto retrieve current data(), you should rather use those functions to measure array dimensions.
  // Also if the scalling factors change (eg drom 2 to 2.5), upon redraw the data of im2 will be rebuild with
  // new pointer and values so be aware that previously retrieved data pointer might not be valid anymore.






  int data_w() const {return image_? image_->w() : 0;}
  int data_h() const {return image_? image_->h() : 0;}

  // These give dimensions of non-scalled original
  int original_w() const;
  int original_h() const;


  // Static ons influencing default behaviour of the class instances


  // This is to register "standard" images so that they can be recognized - equivalent to fl_register_images().
  // but for this new class. This can be removed if/when Fl_Shared_Image is replaced by this class.
  static void register_images();


  // Sets default policy of newly created shared images should they have data available all the time.
  // By switching it off the internal images and data will be created on demand only (usually with first draw)
  // and  data could be deleted afterwards, keeping only cache, saving memory and often also CPU resources.
  // By default it is OFF (0) so that things like copy() and resize() desaturate(), color_average() inactive()
  // are cheap: you might want to switch it ON if you want to access (deprecated) data() function
  // To assure data for particular image function you can just call data() upon it - however be careful:
  // because data() is NOT YET virtual, the pointer variable must be if fl::Fl_Shareable_Image class so that compiler use
  // this version of data - not the one of the Fl_Image base class.
  // These functions should be removed once data() function is declared as virtual.
  static void assure_data_mode(int a);
  static int assure_data_mode();




  // This draws image shrinked or extended obeying transformation matrix.
  // Only scalling is at the moment implemented (rotation/skew x/y mixing coefficients of transformation matrix are disregarded)
  // Note that different copies will share all the data and cache with other copies as long they result to the same
  // "screen" pixels after all transformations - no matter how copies were created.
  // If W or H is 0 it calculates this dimension to keep the aspect ratio. If both are 0 (default) it draws with current w() and h()
  // which are scalled according to current transformation matrix.
  static void draw(Fl_Shareable_Image * im, int X, int Y, int W = 0, int H = 0);



  // You can override usual copy by something better if you wish do do eg better rescalling of images...
  static void copy_function(Fl_Image * (*f)(Fl_Image * im, int W, int H));



};

// Checks if the image is shareble one
inline int fl_image_is_shared(const Fl_Image *im){return ((Fl_Shareable_Image *)im)->base_count() == -1 ? 1 : 0;}

// See  void Fl_Shareable_Image::draw(Fl_Shareable_Image * im, int X, int Y, int W = 0, int H = 0);
inline void fl_draw(Fl_Shareable_Image * im, int X, int Y, int W = 0, int H = 0){ Fl_Shareable_Image::draw(im, X, Y, W, H); }





#endif // _Fl_Sharable_Image_H_
